/*
 * Thanks to Scot Laird, Paul Chinn, Russell Nelson, and others..
 *
 * Copyright (C) 1996 by Thomas Davis
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL SCOTT LAIRD BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#define __KERNEL__
#define MODULE

#include <errno.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <asm/io.h>

#include "quickcam.h"
#include "qcam-kernel.h"

void reset_quickcam(struct quickcam *qc)
{
	outb(0x20, qc->port + 2);
	outb(0x75, qc->port);
	qc->status = inb(qc->port);
	outb(0x0b, qc->port + 2);
	udelay(250);
	outb(0x0e, qc->port + 2);
	udelay(250);
}

static int getstatus(struct quickcam *qc)
{
	unsigned char status1, status2;

	outb(0x06, qc->port + 2);

	qc->count = 255;

	do {
		status1 = inb(qc->port + 1);
	} while (!(status1 & 0x08) && --qc->count);

	if (qc->debug > 1) {
		printk("qc->count = %d\n", qc->count);
	}

	if (qc->count <= 0) {
		printk("quickcam: camera not responding.  check cables.\n");
		return -ENODEV;
	}

	outb(0x0e, qc->port + 2);

	qc->count = 255;
	do {
		status2 = inb(qc->port + 1);
	} while ((status2 & 0x08) && --qc->count) ;

	if (qc->count <= 0) {
		printk("quickcam: camera not responding.  check cables.\n");
		return -ENODEV;
	}

	if (qc->debug > 1) {
		printk("qc->count = %d, ", qc->count);
		printk("status1 = 0x%02x, ", status1);
		printk("status2 = 0x%02x, ", status2);
		printk("status = 0x%02x\n",((status1 & 0xF0) | ( status2 >> 4 )));
	}

	return ((status1 & 0xF0 ) | ( status2 >> 4 ));
}

static int sendbyte(struct quickcam *qc, int value)
{
	outb(value, qc->port);
	outb(value, qc->port);
	outb(value, qc->port);
	return (getstatus(qc));
}

/*
 * sendcmd needs the port, command to send, and the value to send.
 * sends the first value, and the second as pairs of 4 bit words.
 * if the camera is not there, it will error out.
 */

static int sendcmd(struct quickcam *qc, int cmd, int value)
{

	if (qc->debug > 1)
		printk("cmd = 0x%02x, value = 0x%02x\n", cmd, value);

	if (sendbyte(qc, cmd) != cmd )
		return -ENODEV;

	if (sendbyte(qc, value) != value )
		return -ENODEV;

	return 0;
}

/*
 * xfermode table
 *
 * 	X size	4 Bpp	6 Bpp	Mode
 *    ---------------------------------
 *	320	0	2	Uni
 *		1	3	Bi
 *	160	4	6	Uni
 *		5	7	Bi
 *	80	8	10	Uni
 *		9	11	Bi
 *
 * These can be zoom'd also, but that is not in this version of the driver
 * (yet!)
 *
 */

static void set_xfermode(struct quickcam *qc)
{
	if (qc->mode == QC_UNIDIR) {
		qc->xfermode = 0;
	} else {
		qc->xfermode = 1;
	}

	if (qc->bpp == 6)
		qc->xfermode += 2;

	switch (qc->xsize) {
		case 320:
			break;
		case 160:
			qc->xfermode += 4;
			break;
		case 80:
			qc->xfermode += 8;
			break;
	}

	if (qc->debug > 1)
		printk("setting xfermode to %d\n", qc->xfermode);

}

int qcam_find(int port)
{
	struct quickcam qcam, *qc;

	qc = &qcam;
	qc->port = port;
	qc->debug = 0;	

	outb(0x20, qc->port + 2);
	outb(0x75, qc->port);
	qc->status = inb(qc->port);
	outb(0x0b, qc->port + 2);
	udelay(250);
	outb(0x0e, qc->port + 2);
	udelay(250);

	return sendcmd(qc, QC_BRIGHTNESS, 0x01);
}

/*
 * The only thing to rememeber, is if the bits per pixel is 4, then we
 * send xsize / 2.  Otherwise, it's xsize / 4.
 */

int qcam_init(struct quickcam *qc)
{

	set_xfermode(qc);

	if (sendcmd(qc, QC_BRIGHTNESS, qc->brightness) < 0) {
		return -ENODEV;
	}

	sendcmd(qc, QC_BRIGHTNESS, 0x01);
	sendcmd(qc, QC_BRIGHTNESS, 0x01);
	sendcmd(qc, QC_BRIGHTNESS, qc->brightness);
	sendcmd(qc, QC_BRIGHTNESS, qc->brightness);

	sendcmd(qc, QC_UNK1, qc->panx);
	sendcmd(qc, QC_UNK2, qc->pany);
	sendcmd(qc, QC_BRIGHTNESS, qc->brightness);

	if (qc->bpp == 4)
		sendcmd(qc, QC_XSIZE, qc->xsize/2);
	else
		sendcmd(qc, QC_XSIZE, qc->xsize/4);

	sendcmd(qc, QC_YSIZE, qc->ysize);
	sendcmd(qc, QC_CONTRAST, qc->contrast);
	sendcmd(qc, QC_WHITEBALANCE, qc->whitebalance);

	qc->init = 0;

	return 0;
}

int doscan(struct quickcam *qc)
{
	if (qc->init == 1) {
		if (qcam_init(qc) < 0) {
			printk("problems with camera initialization. aborting scan..\n");
			return -ENODEV;
		}
	}

	if (sendcmd(qc, QC_XFERMODE, qc->xfermode) < 0) {
		printk("problems with setting xfermode.. aborting scan..\n");
		return -ENODEV;
	}

	switch (qc->bpp) {
		case 4:
			if (qc->mode == QC_BIDIR && qc->status != 0x75)
				bi_4bit(qc);
			else
				uni_4bit(qc);
			break;
		case 6:
			if (qc->mode == QC_BIDIR && qc->status != 0x75) {
				waitfor_bi(qc);
				bi_6bit(qc);
			} else
				uni_6bit(qc);
			break;
	}

	qc->framecount++;

	if (qc->debug > 1)
		printk("quickcam: finished scan.\n");

	return 0;
}

